# -*- coding: UTF-8 -*-
#
#   ::::::::> awsac-session (08-10-13) <::::::::
#
#   by Jan-Philip Gehrcke (jgehrcke@gmail.com)
#   Universität Würzburg, Max-Planck-Institut für Physik München
# 
#   Copyright (C) 2008 Jan-Philip Gehrcke
#
#   LICENSE:                      
#   This program is free software; you can redistribute it and/or modify 
#   it under the terms of the GNU General Public License as published by 
#   the Free Software Foundation; either version 3 of the License, or 
#   (at your option) any later version. This program is distributed in
#   the hope that it will be useful, but WITHOUT ANY WARRANTY; without
#   even the implied warranty of MERCHANTABILITY or FITNESS FOR A
#   PARTICULAR PURPOSE. See the GNU General Public License for more
#   details. You should have received a copy of the GNU General Public
#   License along with this program; if not, see
#   <http://www.gnu.org/licenses/>.
#
#   read the detailled documentation at http://gehrcke.de/awsac
#
########################################################################

import sys
import boto
import hashlib
import time
import os
import random
import optparse
import ConfigParser

VERSION = "v08-10-13" 

def main():
    """
    lead through the script step by step
    """

    print '\n'
    print '::::>  AWSACtools session management ' + VERSION
    print '::::>  by Jan-Philip Gehrcke'
    print '\n'

    options = parseargs()
    files = check_usergivenfiles(options)

    if options.action == 'start':
        print '# starting new session; parsing start information file... \n'
        s = Session(startinfo=parse_startinfofile(files['startinfofilepath']))
        print '# ----- shortdescription of new session: '+s.shortdescr.upper()\
               +' -----\n'
        
        s.calc_n_instances()
        print ('# instancetype chosen: `'+s.instance_type+'` with '+s.cores_per_instance+' '
               'core(s) per instance. With '+s.n_jobs+' demanded job(s)/core(s) '
               'this makes '+s.n_instances+' instance(s) we have to start. '
               'Unused cores: '+s.unused_cores+'.\n')
               
        print '# generating session ID from date, shortdescription, randomness:\n',
        s.gen_id()
        print s.session_id+'\n'
        
        print '# uploading sessionarchive file to S3...'
        if s.s3_upload_file_to_sessionfolder(files['archivefilepath']):
            print 'uploaded to S3 as object (bucket:`'+s.bucket+'` key:`'+s.session_id+'/'\
                                  +os.path.basename(files['archivefilepath'])+'`).\n'

        #s.s3_upload_datadir_to_sessionfolder(files['datadirpath']

        print '# connecting to EC2 to check your AMI-ID...'
        s.check_image()
        print 'found AMI `'+s.ami.id+'` ('+s.ami.location+')\n'

        print '# building user-data string to be submitted to all instances...'
        s.build_runuserdata(files['archivefilepath'])
        print s.runuserdata+'\n(session_id;bucket;sessionarchivename;cores_per_instance)\n'

        print '# checking SimpleDB domain for this session...'
        s.sdb_delete_domain()

        print '\n# In the next step EC2 will be instructed to run exactly '\
               +s.n_instances+' instance(s) of given AMI (type `'+s.instance_type\
               +'`) with user-data mentioned obove.'
        if get_y_n("proceed? (y/n): ", ['y','n']) == "n":
            sys.exit('Then check your settings again (-: exiting.')
        if s.ec2_runinstances():
            print 'Request accepted. EC2 reservation ID: '+str(s.ec2_reservation.id)

        print '# saving session to file...',
        print s.save_to_file()

    if options.action == 'check':
        cfg = parse_configfile(files['configfilepath'])
        read_sessionstate_from_sdb(cfg['session_id'])
    
    if options.action == 'getresults':
        cfg = parse_configfile(files['configfilepath'])
        s3_download_files(bucketname=cfg['bucket'],
                          prefix=cfg['session_id'],
                          outputdir=os.path.join(files['outputdirpath'],'session-'+cfg['session_id']))

    if options.action == 'cleanup':
        cfg = parse_configfile(files['configfilepath'])
        deldomain = get_y_n('delete SimpleDB-domain '+cfg['session_id']+'? y/n: ', ['y','n'])
        if deldomain == 'y':
            sdb_delete_domain(cfg['session_id'])
        dels3keys = get_y_n('delete S3-objects '+cfg['bucket']+'/'+cfg['session_id']+'/* ? y/n: ', ['y','n'])
        if dels3keys == 'y':
            s3_delete_keys(bucketname=cfg['bucket'],
                           prefix=cfg['session_id'])


class Session:
    """
    Represents and manages one session.
    """

    def __init__(self, startinfo):
        self.session_id = None
        self.n_instances = None
        self.unused_cores = None
        self.cores_per_instance = None
        self.ami = None # will be a boto image object
        self.runuserdata = None
        self.ec2_reservation = None
        self.shortdescr=startinfo['shdescr']
        self.bucket=startinfo['bckt']
        self.n_jobs=startinfo['n_jobs']
        self.ami_id=startinfo['ami_id']
        self.ec2_uid=startinfo['ec2_uid']
        self.instance_type=startinfo['instance_type']


    def build_runuserdata(self, archivefile):
        """
        Build the user-data string and save it in attribute `runuserdata`

        user-data must have following form:
        session_id;bucket;sessionarchivename;cores_per_instance
        """
        arcfilename = os.path.basename(archivefile)
        self.runuserdata = self.session_id+';'+self.bucket+';'+\
                           arcfilename+';'+self.cores_per_instance

    def calc_n_instances(self):
        """
        Calculate the number of EC2-instances in this session from n_jobs and n_cpi,
        the number of cores per instance.

        n_cpi is obtained from the instance-type.
        The number of unused cores is calculated, too (e.g. 3 jobs on c1.medium lead
        to one unused core).
        """
        if self.instance_type == 'm1.small':
            self.cores_per_instance = str(1)
        elif self.instance_type == 'c1.medium':
            self.cores_per_instance = str(2)
        n_jobs = int(self.n_jobs)
        cores_per_instance = int(self.cores_per_instance)
        n_instances = n_jobs / cores_per_instance
        overhead = n_jobs % cores_per_instance
        unused_cores = 0
        if overhead:
             n_instances += 1
             unused_cores = cores_per_instance - overhead
        self.n_instances = str(n_instances)
        self.unused_cores = str(unused_cores)

    def check_image(self):
        """
        Check if the usergiven amiID represents an existing Amazon Machine Image.

        If not, the script will quit.
        """
        ownerlist = []
        ownerlist.append(self.ec2_uid)
        myimage = None
        conn = boto.connect_ec2()
        myimages=conn.get_all_images(owners=ownerlist)
        for image in myimages:
            if image.id == self.ami_id:
                myimage = image
                break
        if myimage is None:
            sys.exit(("desired AMI `"+self.ami_id+"` was not in the list of"
                      " EC2-user's (UID: "+self.ec2_uid+") AMIs. exiting."))
        self.ami = myimage

    def ec2_runinstances(self):
        """
        Invoke the 'RunInstances" API-call using boto.

        @return: True. But: an error while API-call would result in exit.
        """
        conn = boto.connect_ec2()
        reservation=conn.run_instances(min_count=self.n_instances,
                                       max_count=self.n_instances,
                                       user_data=self.runuserdata,
                                       image_id=self.ami_id,
                                       instance_type=self.instance_type)
        self.ec2_reservation = reservation
        return True

    def save_to_file(self):
        """
        Save current session information to a file using ConfigParser.
        Put session id in filename.
        """
        filename = 'session-'+self.session_id+'.cfg'
        try:
            config = ConfigParser.SafeConfigParser()
            config.add_section('sessionconfig')
            config.set('sessionconfig', 'session_id', self.session_id)
            config.set('sessionconfig', 'ec2_reservation_id', str(self.ec2_reservation.id))
            config.set('sessionconfig', 'n_instances', str(self.n_instances))
            config.set('sessionconfig', 'unused_cores', str(self.unused_cores))
            config.set('sessionconfig', 'cores_per_instance', str(self.cores_per_instance))
            config.set('sessionconfig', 'runuserdata', self.runuserdata)
            config.set('sessionconfig', 'shortdescr', self.shortdescr)
            config.set('sessionconfig', 'bucket', self.bucket)
            config.set('sessionconfig', 'n_jobs', str(self.n_jobs))
            config.set('sessionconfig', 'ami_id', self.ami_id)
            config.set('sessionconfig', 'ec2_uid', str(self.ec2_uid))
            config.set('sessionconfig', 'instance_type', self.instance_type)
            fd = open(filename,'w')
            config.write(fd)
            fd.close()
        except:
            return 'error: could not write configfile: '+filename
        return filename

    def gen_id(self):
        """
        Generate session id from current time, shortdescription and randomhash
        """
        try: urandom = os.urandom(20)
        except: urandom = time.time()
        rnd = hashlib.sha1(str(random.random())+str(time.time())+str(urandom))
        sessionsuffix = rnd.hexdigest()[:4]
        timestring = time.strftime("%y%m%d_%H%M",time.localtime())
        self.session_id = timestring+"--"+self.shortdescr[:8]+"--"+sessionsuffix
        #self.session_id = "080910_1147--testsess--609f"

    def sdb_delete_domain(self):
        """
        Delete SimpleDB domain of current session
        """
        sdb_delete_domain(self.session_id)

    def s3_upload_file_to_sessionfolder(self, file):
        """
        Upload a file as S3 object to sessionsbucket/sessionid/file.
        
        @return: True (boto raises exceptions if something goes wrong)
        """
        conn = boto.connect_s3()
        bucket = conn.create_bucket(self.bucket)
        k = boto.s3.key.Key(bucket)
        k.key=self.session_id+"/"+os.path.basename(file)
        k.set_contents_from_filename(file)
        return True

    def s3_download_file_from_sessionfolder(self, getfile, outfile):
        """
        Download a file from S3 object sessionsbucket/sessionid/getfile
        to outfile.
        """
        conn = boto.connect_s3()
        bucket = conn.create_bucket(self.bucket)
        k = boto.s3.key.Key(bucket)
        k.key=self.session_id+"/"+getfile
        k.get_contents_to_filename(outfile)


def s3_delete_keys(bucketname, prefix):
    """
    Delete S3 objects beginning with `prefix` from `bucketname`.
    """
    conn = boto.connect_s3()
    bucketfound = False
    try:
        buckets = conn.get_all_buckets()
    except:
        print '  error while retrieving all buckets'
    if buckets:
        for bucket in buckets:
            if bucket.name == bucketname:
                bucketfound = True
                break
    if not bucketfound:
        print '  bucket could not be found: '+bucketname
    else:
        objectkeys = bucket.list(prefix=prefix)
        for key in objectkeys:
            key.delete()
            print 'deleted '+os.path.basename(key.name)


def sdb_delete_domain(domainname):
    """
    Delete SimpleDB domain `domainname`.
    """
    sdb = boto.connect_sdb()
    domains = ''
    try:
        domains = sdb.get_all_domains()
    except:
        print '  error while retrieving all domains'
    if domains:
        for domain in domains:
            if domain.name == domainname:
                sdb.delete_domain(domain)
                print '  domain existed and was deleted: '+domainname
                return
    print '  domain does not exist: '+domainname


def read_sessionstate_from_sdb(session_id):
    """
    Read SimpleDB domain `session_id` content and print it.
    
    Some special keynames are hardcoded in this function.
    """
    sdb = boto.connect_sdb()
    domainfound = False
    try:
        domains = sdb.get_all_domains()
    except:
        print '  error while retrieving all SimpleDB-domains'
    if domains:
        for domain in domains:
            if domain.name == session_id:
                domainfound = True
                break
    if not domainfound:
        print '  domain for this session could not be found: '+session_id
    else:
        # jobdict_list will be a list of jobdicts
        jobdict_list = []
        for jobitem in domain:
            # store all the key/value-pairs of the db-item in `jobdict`
            jobdict={}
            for key in jobitem.keys():
                jobdict[key] = jobitem[key]
            jobdict_list.append(jobdict)
        if jobdict_list:
            # lets sort the dictslist by a specific key in the dicts
            # I assume that the following keys exist but of course I will check it:
            # jobnr, status, instance_id, launchindex, shellscript, runstarttime,
            # runendtime, returncode, savestarttime, saveendtime
            jobdict_list.sort(key= lambda x: x['jobnr'])
            for jobdict in jobdict_list:
                try:
                    print '============= Job '+jobdict['jobnr']+' (status: '+jobdict['status']+') =============='
                except KeyError:
                    pass
                try:
                    print 'running '+jobdict['shellscript']+' on instance '+\
                          jobdict['instance_id']+' with launchindex '+jobdict['launchindex']
                except KeyError:
                    pass
                try:
                    print 'started running: '+jobdict['runstarttime']
                except KeyError:
                    pass
                try:
                    print 'ended running: '+jobdict['runendtime']
                except KeyError:
                    pass
                try:
                    print 'returncode: '+jobdict['returncode']
                except KeyError:
                    pass
                try:
                    print 'started saving: '+jobdict['savestarttime']
                except KeyError:
                    pass
                try:
                    print 'ended saving: '+jobdict['saveendtime']
                except KeyError:
                    pass
                print ''


def get_y_n(msg, commandlist):
    """
    Get user input from stdin using raw_input() with `msg` as prompt.
    
    Wait for any string that is in stringlist `commandlist`.
    @return: entered string (that is in `commandlist`)
    """
    while True: # wait for instring to be a known command
        while True: # wait for an errorfree input
            try:
                instring = raw_input(msg)
                break
            except:
                pass
        if instring in commandlist:
            return instring


def s3_download_files(bucketname, prefix, outputdir):
    """
    Download files from S3 objects `bucketname`/`prefix`* to `outputdir`.
    
    At first receive all buckets. Then get all objectkeys in `bucketname` with
    specified `prefix`. Download objects to `outputdir`. Ask for overwriting.
    """
    conn = boto.connect_s3()
    bucketfound = False
    try:
        buckets = conn.get_all_buckets()
    except:
        print '  error while retrieving all buckets'
    if buckets:
        for bucket in buckets:
            if bucket.name == bucketname:
                bucketfound = True
                break
    if not bucketfound:
        print '  bucket could not be found: '+bucketname
    else:
        try:
            outputdir = os.path.abspath(outputdir)
            os.makedirs(outputdir)
            print "outputfolder created: "+outputdir
        except OSError:
            print "outputfolder exists: "+outputdir
        overwriteall = False
        objectkeys = bucket.list(prefix=prefix)
        for key in objectkeys:
            outfilepath = os.path.join(outputdir,os.path.basename(key.name))
            if os.path.exists(outfilepath) and not overwriteall:
                instring = get_y_n(os.path.basename(key.name)+' exists. overwrite? y/n/a: ',['y','n','a'])
                if instring == 'n':
                    continue
                if instring == 'a':
                    overwriteall = True
            key.get_contents_to_filename(outfilepath)
            print 'saved '+os.path.basename(key.name)


def parse_configfile(file):
    """
    Parse given `file`name for session information/configuration.
    
    @return: return dict with special options if filecontent is as expected.
    """
    config = ConfigParser.SafeConfigParser()
    try:
        config.readfp(open(file))
        sesscfg = {}
        sesscfg['session_id'] = config.get('sessionconfig', 'session_id')
        sesscfg['ec2_reservation_id'] = config.get('sessionconfig', 'ec2_reservation_id')
        sesscfg['n_instances'] = config.get('sessionconfig', 'n_instances')
        sesscfg['unused_cores'] = config.get('sessionconfig', 'unused_cores')
        sesscfg['cores_per_instance'] = config.get('sessionconfig', 'cores_per_instance')
        sesscfg['runuserdata'] = config.get('sessionconfig', 'runuserdata')
        sesscfg['shortdescr'] = config.get('sessionconfig', 'shortdescr')
        sesscfg['bucket'] = config.get('sessionconfig', 'bucket')
        sesscfg['n_jobs'] = config.get('sessionconfig', 'n_jobs')
        sesscfg['ami_id'] = config.get('sessionconfig', 'ami_id')
        sesscfg['ec2_uid'] = config.get('sessionconfig', 'ec2_uid')
        sesscfg['instance_type'] = config.get('sessionconfig', 'instance_type')
        config.write(sys.stdout)
    except:
        sys.exit(('error while parsing '+file+' for configuration options.\n'
                  'expecting section [startinfo] with following options:\n'
                  'session_id,ec2_reservation_id,n_instances,unused_cores,cores_per_instance,\n'
                  'runuserdata,shortdescr,bucket,n_jobs,ami_id,ec2_uid,instance_type'))
    return sesscfg



def parse_startinfofile(file):
    """
    Check configuration file for options.

    Write the parsed data to stdout.
    @return: information as strings without whitespaces
    """
    config = ConfigParser.SafeConfigParser()
    try:
        config.readfp(open(file))
        sessinfo = {}
        sessinfo['shdescr'] = ''.join(config.get('startinfo','shortdescr').split())
        sessinfo['n_jobs'] = ''.join(config.get('startinfo','n_jobs').split())
        sessinfo['bckt'] = ''.join(config.get('startinfo','sessionsbucket').split())
        sessinfo['ami_id'] = ''.join(config.get('startinfo','ami_id').split())
        sessinfo['ec2_uid'] = ''.join(config.get('startinfo','ec2_uid').split())
        sessinfo['instance_type'] = ''.join(config.get('startinfo','instance_type').split())
        config.write(sys.stdout)
    except:
        sys.exit(('error while parsing '+file+' for configuration options.\n'
                  'expecting section [startinfo] with following options:\n'
                  'shortdescr,n_jobs,sessionsbucket,ami_id,ec2_uid,instance_type'))
    return sessinfo


# checks if the script really can work with user-given options.
# logical consistence has already been checked before, so lets look if
# the given files and directories exist the way we need it.
# returns dictionary of files and directories with absolute pathes.
def check_usergivenfiles(options):
    """
    Check if the files given by commandline exist. `check_file`-fkt will raise errors.

    @return: A dictionary containing the absolute paths of the files.
    """
    archivefilepath = None
    configfilepath = None
    startinfofilepath = None
    outputdirpath = None
    datadirpath = None
    if options.action == 'start':
        archivefilepath = check_file(options.archivefile)
        if not os.path.basename(archivefilepath).endswith("tar.bz2"):
            sys.exit(("The session archive file has to be a bz2 compressed tarball. The file "
                      "you submitted does not end with 'tar.bz2'. So I suppose wrong content."))
        startinfofilepath = check_file(options.startinfofile)
        if options.datadir is not None:
            datadirpath = check_dir(options.datadir)
    if options.action == 'check':
        configfilepath = check_file(options.configfile)
    if options.action == 'getresults':
        outputdirpath = check_dir(options.outputdir)
        configfilepath = check_file(options.configfile)
    if options.action == 'cleanup':
        configfilepath = check_file(options.configfile)
    return {'archivefilepath': archivefilepath,
            'configfilepath': configfilepath,
            'startinfofilepath': startinfofilepath,
            'outputdirpath': outputdirpath,
            'datadirpath': datadirpath}


def parseargs():
    """
    Parse commandlineoptions using the optparse module and check them for the
    logical consistence. Generate help- and usage-output.

    @return: optparse `options`-object containing the commandlineoptions
    """
    # the following part configures the OptionParser...
    version = 'AWSACtools session management' + VERSION
    description = ("AWSAC session management" + VERSION + " - start sessions, check status "
                   "of sessions, get result data and clean up used AWS")
    usage = ("%prog [--start || --check || --getresults || --cleanup]\n\t\t"
             "[-a A -i I (-d D) || -c C || -c C -o O || -c C]\n\t\t"
             "type -h, --help for help and --version for versionoutput")
    dhelp = ('directory to grab additional session data from. this data will be made'
             ' available to the instances by uploading it to S3.'
             ' (-d is optional for `start`) WARNING: CURRENTLY NOT SUPPORTED')
    parser = optparse.OptionParser(usage=usage,version=version,description=description)
    parser.add_option('-a', '--archive', dest='archivefile',
                      help='session archive file (needed for session `start`)')
    parser.add_option('-i', '--ini', dest='startinfofile',
                      help='session startinformation file (needed for session `start`)')
    parser.add_option('-d', '--data', dest='datadir',
                      help=dhelp)
    parser.add_option('-c', '--config', dest='configfile',
                      help='session config file (needed for `check`, `getresults`, `cleanup`)')
    parser.add_option('-o', '--outdir', dest='outputdir',
                      help='directory to save results in (needed for `getresults`; dir must exist - subfolder with sessionname will be created)')
    parser.add_option('--start', action='store_const', const='start', dest='action',
                      help='start new session (needs ARCHIVEFILE and STARTINFOFILE set - may have DATADIR set)')
    parser.add_option('--check', action='store_const', const='check', dest='action',
                      help='check status of an existing session (needs CONFIGFILE set)')
    parser.add_option('--getresults', action='store_const', const='getresults', dest='action',
                      help=('get result data of an existing session '
                            '(needs CONFIGFILE and OUTPUTDIR set)'))
    parser.add_option('--cleanup', action='store_const', const='cleanup', dest='action',
                      help='deletes all s3 objects of an existing session (needs CONFIGFILE set)')

    # now read in the given arguments (from sys.argv by default)
    (options, args) = parser.parse_args()

    # now check the logical consistence...
    if options.action is None:
        parser.error('one of [--start || --check || --getresults || --cleanup] must be set!')
    elif options.action == 'start':
        if (options.archivefile is None) or (options.startinfofile is None):
            parser.error('when --start is set, -a ARCHIVEFILE and -i STARTINFOFILE are needed!')
    elif options.action == 'check':
        if (options.configfile is None):
            parser.error('when --check is set, -c CONFIGFILE is needed!')
    elif options.action == 'getresults':
        if (options.configfile is None) or (options.outputdir is None):
            parser.error('when --getresults is set, -c CONFIGFILE and -o OUTPUTDIR are needed!')
    elif options.action == 'cleanup':
        if (options.configfile is None):
            parser.error('when --cleanup is set, -c CONFIGFILE is needed!')

    # everything okay until here? seems so.. return the given options.
    return options


def check_file(file):
    """
    Check if a given file exists and really is a file (e.g. not a directory)
    In errorcase the script is stopped.

    @return: the absolute path of the file
    """
    if not os.path.exists(file):
        sys.exit(file+' does not exist')
    if not os.path.isfile(file):
        sys.exit(file+' must be a file')
    return os.path.abspath(file)


def check_dir(dir):
    """
    Check if a given dir exists and really is a dir (e.g. not a file)
    In errorcase the script is stopped.

    @return: the absolute path of the directory
    """
    if not os.path.exists(dir):
        sys.exit(dir+' does not exist')
    if not os.path.isdir(dir):
        sys.exit(dir+' is not a directory')
    return os.path.abspath(dir)


if __name__ == "__main__":
    main()
